import logging
from typing import BinaryIO

import boto3
from botocore.client import Config
from botocore.exceptions import ClientError

from udata_hydra import config

log = logging.getLogger("udata-hydra")


def get_resource_minio_url(key: str, resource_id: str) -> str:
    """Returns location of given resource in minio once it is saved"""
    return f"{config.MINIO_URL}/{config.MINIO_BUCKET}/{config.MINIO_FOLDER}/{key}/{resource_id}"


def save_resource_to_minio(resource_file: BinaryIO, key: str, resource_id: str) -> None:
    log.info("Saving to minio")
    log.debug(f"Bucket is {config.MINIO_BUCKET}")
    s3 = boto3.client(
        "s3",
        endpoint_url=config.MINIO_URL,
        aws_access_key_id=config.MINIO_USER,
        aws_secret_access_key=config.MINIO_PWD,
        config=Config(signature_version="s3v4"),
    )
    try:
        with open(resource_file.name, "rb") as f:
            s3.upload_fileobj(
                Fileobj=f,
                Bucket=config.MINIO_BUCKET,
                Key=f"{config.MINIO_FOLDER}/{key}/{resource_id}",
            )
        log.info(f"Resource saved into minio at {get_resource_minio_url(key, resource_id)}")
    except ClientError as e:
        log.error(e)


def delete_resource_from_minio(key: str, resource_id: str) -> None:
    log.info("Deleting from minio")
    s3 = boto3.client(
        "s3",
        endpoint_url=config.MINIO_URL,
        aws_access_key_id=config.MINIO_USER,
        aws_secret_access_key=config.MINIO_PWD,
        config=Config(signature_version="s3v4"),
    )
    try:
        s3.delete_object(
            Bucket=config.MINIO_BUCKET,
            Key=f"{config.MINIO_FOLDER}/{key}/{resource_id}",
        )
        log.info(f"Resource deleted from minio at {get_resource_minio_url(key, resource_id)}")
    except ClientError as e:
        log.error(e)



from marshmallow import Schema, fields


class ResourceDocumentSchema(Schema):
    id = fields.Str(required=True)
    url = fields.Str(required=True)
    format = fields.Str(allow_none=True)
    title = fields.Str(required=True)
    schema = fields.Dict(allow_none=True)
    description = fields.Str(allow_none=True)
    filetype = fields.Str(required=True)
    type = fields.Str(required=True)
    mime = fields.Str(allow_none=True)
    filesize = fields.Int(allow_none=True)
    checksum_type = fields.Str(allow_none=True)
    checksum_value = fields.Str(allow_none=True)
    created_at = fields.DateTime(required=True)
    last_modified = fields.DateTime(required=True)
    extras = fields.Dict()
    harvest = fields.Dict()


class ResourceSchema(Schema):
    dataset_id = fields.Str(required=True)
    resource_id = fields.Str(required=True)
    document = fields.Nested(ResourceDocumentSchema(), allow_none=True)



import logging
from typing import Union
from unittest.mock import MagicMock

import asyncpg
import redis
from rq import Queue

from udata_hydra import config

log = logging.getLogger("udata-hydra")
context = {
    "databases": {},
    "queues": {},
}


def monitor() -> MagicMock:
    if "monitor" in context:
        return context["monitor"]
    monitor = MagicMock()
    monitor.set_status = lambda x: log.debug(x)
    monitor.init = lambda **kwargs: log.debug(f"Starting udata-hydra... {kwargs}")
    context["monitor"] = monitor
    return context["monitor"]


async def pool(db: str = "main") -> asyncpg.pool.Pool:
    if db not in context["databases"]:
        dsn = config.DATABASE_URL if db == "main" else getattr(config, f"DATABASE_URL_{db.upper()}")
        context["databases"][db] = await asyncpg.create_pool(
            dsn=dsn,
            max_size=config.MAX_POOL_SIZE,
            server_settings={"search_path": config.DATABASE_SCHEMA},
        )
    return context["databases"][db]


def queue(name: str = "default") -> Union[Queue, None]:
    if not context["queues"].get(name):
        # we dont need a queue while testing, make sure we're not using a real Redis connection
        if config.TESTING:
            return None
        connection = redis.from_url(config.REDIS_URL)
        context["queues"][name] = Queue(
            name, connection=connection, default_timeout=config.RQ_DEFAULT_TIMEOUT
        )
    return context["queues"][name]



import logging
from pathlib import Path

from asyncpg.exceptions import SyntaxOrAccessError

from udata_hydra import context

log = logging.getLogger("udata-hydra")


class Migrator:
    # NB: we can't use async __init__, hence the workaround
    @classmethod
    async def create(cls, db_name, skip_errors=False) -> "Migrator":
        self = Migrator(db_name)
        self.skip_errors = skip_errors
        self.db = await context.pool(db=db_name)
        q = f"""CREATE TABLE IF NOT EXISTS {self.table_name}(
            id serial PRIMARY KEY,
            name VARCHAR,
            status VARCHAR,
            created_at TIMESTAMP DEFAULT NOW()
        );"""
        await self.db.execute(q)
        return self

    def __init__(self, db_name) -> None:
        self.db_name = db_name
        self.table_name = f"migrations_{db_name}"

    async def get(self, name) -> dict:
        q = f"SELECT * FROM {self.table_name} WHERE name = $1 AND status = $2"
        return await self.db.fetchrow(q, name, "DONE")

    async def register(self, name) -> dict:
        q = f"INSERT INTO {self.table_name}(name, status) VALUES($1, $2)"
        await self.db.execute(q, name, "DONE")

    async def apply(self, migration_file: Path):
        name = migration_file.stem
        existing = await self.get(name)
        if not existing:
            log.info(f"Applying {name}")
            with migration_file.open("r") as f:
                try:
                    await self.db.execute(f.read())
                except SyntaxOrAccessError as e:
                    if not self.skip_errors:
                        raise e
                    log.warning(f"Error while applying {name} ({e}), skipping and registering")
            await self.register(name)
        else:
            log.debug(f"Skipping {name}, already applied at {existing['created_at']}")

    async def migrate(self):
        migrations_path = Path(__file__).parent / self.db_name
        # Files like 20230130_migration-name.sql
        mfiles = list(migrations_path.glob("????????_*.sql"))
        mfiles.sort(key=lambda x: x.stem)
        for m in mfiles:
            await self.apply(m)



import logging
from typing import Union

import toml

log = logging.getLogger("udata-hydra")


def get_app_version() -> Union[str, None]:
    """Get the app version from pyproject.toml"""
    try:
        pyproject: dict = toml.load("pyproject.toml")
        app_version: str = pyproject["tool"]["poetry"]["version"]
        return app_version
    except Exception as e:
        log.error(f"Error while getting app version: {e}")
        return None



import json

from aiohttp import web
from marshmallow import ValidationError

from udata_hydra import config
from udata_hydra.db.resource import Resource
from udata_hydra.schemas import ResourceSchema
from udata_hydra.utils import delete_resource_from_minio, get_request_params


async def get_resource(request: web.Request) -> web.Response:
    """Endpoint to get a resource from the DB
    Respond with a 200 status code and a JSON body with the resource data
    If resource is not found, respond with a 404 status code
    """
    [resource_id] = get_request_params(request, params_names=["resource_id"])
    resource = await Resource.get(resource_id)
    if not resource:
        raise web.HTTPNotFound()

    return web.json_response(ResourceSchema().dump(dict(resource)))


async def create_resource(request: web.Request) -> web.Response:
    """Endpoint to receive a resource creation event from a source
    Will create a new resource in the DB "catalog" table and mark it as priority for next crawling
    Respond with a 200 status code and a JSON body with a message key set to "created"
    If error, respond with a 400 status code
    """
    try:
        payload = await request.json()
        valid_payload: dict = ResourceSchema().load(payload)
    except ValidationError as err:
        raise web.HTTPBadRequest(text=json.dumps(err.messages))

    resource = valid_payload["document"]
    if not resource:
        raise web.HTTPBadRequest(text="Missing document body")

    dataset_id = valid_payload["dataset_id"]
    resource_id = valid_payload["resource_id"]

    await Resource.insert(
        dataset_id=dataset_id,
        resource_id=resource_id,
        url=resource["url"],
        priority=True,
    )

    return web.json_response({"message": "created"})


async def update_resource(request: web.Request) -> web.Response:
    """Endpoint to receive a resource update event from a source
    Will update an existing resource in the DB "catalog" table and mark it as priority for next crawling
    Respond with a 200 status code and a JSON body with a message key set to "updated"
    If error, respond with a 400 status code
    """
    try:
        payload = await request.json()
        valid_payload: dict = ResourceSchema().load(payload)
    except ValidationError as err:
        raise web.HTTPBadRequest(text=json.dumps(err.messages))

    resource = valid_payload["document"]
    if not resource:
        raise web.HTTPBadRequest(text="Missing document body")

    dataset_id = valid_payload["dataset_id"]
    resource_id = valid_payload["resource_id"]

    await Resource.update_or_insert(dataset_id, resource_id, resource["url"])

    return web.json_response({"message": "updated"})


async def delete_resource(request: web.Request) -> web.Response:
    try:
        payload = await request.json()
        valid_payload: dict = ResourceSchema().load(payload)
    except ValidationError as err:
        raise web.HTTPBadRequest(text=json.dumps(err.messages))

    dataset_id = valid_payload["dataset_id"]
    resource_id = valid_payload["resource_id"]

    pool = request.app["pool"]
    async with pool.acquire() as connection:
        if config.SAVE_TO_MINIO:
            delete_resource_from_minio(dataset_id, resource_id)
        # Mark resource as deleted in catalog table
        q = f"""UPDATE catalog SET deleted = TRUE WHERE resource_id = '{resource_id}';"""
        await connection.execute(q)

    return web.json_response({"message": "deleted"})



import json
import logging
import os
from datetime import datetime, timezone
from enum import Enum
from typing import Tuple, Union

import magic
from dateparser import parse as date_parser

from udata_hydra import config, context
from udata_hydra.analysis.csv import analyse_csv
from udata_hydra.db.check import Check
from udata_hydra.utils import (
    compute_checksum_from_file,
    detect_tabular_from_headers,
    download_resource,
    queue,
    send,
)


class Change(Enum):
    HAS_CHANGED = 1
    HAS_NOT_CHANGED = 2
    NO_GUESS = 3


log = logging.getLogger("udata-hydra")


async def process_resource(check_id: int, is_first_check: bool) -> None:
    """
    Perform analysis on the resource designated by check_id
    - change analysis
    - size (optionnal)
    - mime_type (optionnal)
    - checksum (optionnal)
    - launch csv_analysis if looks like a CSV response

    Will call udata if first check or changes found, and update check with optionnal infos
    """
    check: dict = await Check.get(check_id)
    if not check:
        log.error(f"Check not found by id {check_id}")
        return

    exceptions = config.LARGE_RESOURCES_EXCEPTIONS

    resource_id = check["resource_id"]
    dataset_id = check["dataset_id"]
    url = check["url"]
    headers = json.loads(check["headers"] or "{}")

    log.debug(f"Analysis for resource {resource_id} in dataset {dataset_id}")

    # let's see if we can infer a modification date on early hints based on harvest infos and headers
    change_status, change_payload = await detect_resource_change_on_early_hints(resource_id)

    # could it be a CSV? If we get hints, we will analyse the file further depending on change status
    is_tabular, file_format = await detect_tabular_from_headers(check)
    max_size_allowed = (
        None if str(resource_id) in exceptions else float(config.MAX_FILESIZE_ALLOWED[file_format])
    )

    # if the change status is NO_GUESS or HAS_CHANGED, let's download the file to get more infos
    dl_analysis = {}
    tmp_file = None
    if change_status != Change.HAS_NOT_CHANGED:
        try:
            tmp_file = await download_resource(url, headers, max_size_allowed)
        except IOError:
            dl_analysis["analysis:error"] = "File too large to download"
        else:
            # Get file size
            dl_analysis["analysis:content-length"] = os.path.getsize(tmp_file.name)
            # Get checksum
            dl_analysis["analysis:checksum"] = compute_checksum_from_file(tmp_file.name)
            # Check if checksum has been modified if we don't have other hints
            if change_status == Change.NO_GUESS:
                (
                    change_status,
                    change_payload,
                ) = await detect_resource_change_from_checksum(
                    resource_id, dl_analysis["analysis:checksum"]
                )
            dl_analysis["analysis:mime-type"] = magic.from_file(tmp_file.name, mime=True)
        finally:
            if tmp_file and not is_tabular:
                os.remove(tmp_file.name)
            await Check.update(
                check_id,
                {
                    "checksum": dl_analysis.get("analysis:checksum"),
                    "analysis_error": dl_analysis.get("analysis:error"),
                    "filesize": dl_analysis.get("analysis:content-length"),
                    "mime_type": dl_analysis.get("analysis:mime-type"),
                },
            )

    if change_status == Change.HAS_CHANGED:
        await store_last_modified_date(change_payload or {}, check_id)

    analysis_results = {**dl_analysis, **(change_payload or {})}
    if change_status == Change.HAS_CHANGED or is_first_check:
        if is_tabular and tmp_file:
            queue.enqueue(analyse_csv, check_id, file_path=tmp_file.name, _priority="default")
        queue.enqueue(
            send,
            dataset_id=dataset_id,
            resource_id=resource_id,
            document=analysis_results,
            _priority="high",
        )


async def store_last_modified_date(change_analysis: dict, check_id: int) -> None:
    """
    Store last modified date in checks because it may be useful for later comparison
    """
    last_modified = change_analysis.get("analysis:last-modified-at")
    if last_modified:
        last_modified = datetime.fromisoformat(last_modified)
        await Check.update(check_id, {"detected_last_modified_at": last_modified})


async def detect_resource_change_from_checksum(
    resource_id, new_checksum
) -> Tuple[Change, Union[dict, None]]:
    """
    Checks if resource checksum has changed over time
    Returns a tuple with a Change status and an optional payload:
    {
        "analysis:last-modified-at": last_modified_date,
        "analysis:last-modified-detection": "computed-checksum",
    }
    """
    q = """
    SELECT checksum
    FROM checks
    WHERE resource_id = $1 and checksum IS NOT NULL
    ORDER BY created_at DESC
    LIMIT 1
    """
    pool = await context.pool()
    async with pool.acquire() as connection:
        data = await connection.fetchrow(q, resource_id)
        if data and data["checksum"] != new_checksum:
            return Change.HAS_CHANGED, {
                "analysis:last-modified-at": datetime.now(timezone.utc).isoformat(),
                "analysis:last-modified-detection": "computed-checksum",
            }
    return Change.NO_GUESS, None


async def detect_resource_change_from_last_modified_header(
    data: dict,
) -> Tuple[Change, Union[dict, None]]:
    # last modified header check

    if len(data) == 1 and data[0]["last_modified"]:
        last_modified_date = date_parser(data[0]["last_modified"])
        return Change.HAS_CHANGED, {
            "analysis:last-modified-at": last_modified_date.isoformat(),
            "analysis:last-modified-detection": "last-modified-header",
        }

    if len(data) == 1 or not data[0]["last_modified"]:
        return Change.NO_GUESS, None

    if data[0]["last_modified"] != data[1]["last_modified"]:
        last_modified_date = date_parser(data[0]["last_modified"])
        return Change.HAS_CHANGED, {
            "analysis:last-modified-at": last_modified_date.isoformat(),
            "analysis:last-modified-detection": "last-modified-header",
        }
    return Change.HAS_NOT_CHANGED, None


async def detect_resource_change_from_content_length_header(
    data: dict,
) -> Tuple[Change, Union[dict, None]]:
    # content-length variation between current and last check
    if len(data) <= 1 or not data[0]["content_length"]:
        return Change.NO_GUESS, None
    if data[0]["content_length"] != data[1]["content_length"]:
        changed_at = data[0]["created_at"]
        return Change.HAS_CHANGED, {
            "analysis:last-modified-at": changed_at.isoformat(),
            "analysis:last-modified-detection": "content-length-header",
        }
    return Change.HAS_NOT_CHANGED, None


async def detect_resource_change_on_early_hints(
    resource_id: str,
) -> Tuple[Change, Union[dict, None]]:
    """
    Try to guess if a resource has been modified from harvest and headers in check data:
    - last-modified header value if it can be found and parsed
    - content-length if it is found and changed over time (vs last checks)

    Returns a tuple with a Change status and an optional payload:
    {
        "analysis:last-modified-at": last_modified_date,
        "analysis:last-modified-detection": "detection-method",
    }
    """
    # Fetch current and last check headers
    q = """
    SELECT
        created_at,
        headers->>'last-modified' as last_modified,
        headers->>'content-length' as content_length,
        detected_last_modified_at
    FROM checks
    WHERE resource_id = $1
    ORDER BY created_at DESC
    LIMIT 2
    """
    pool = await context.pool()
    async with pool.acquire() as connection:
        data = await connection.fetch(q, resource_id)

    # not enough checks to make a comparison
    if not data:
        return Change.NO_GUESS, None

    # let's see if we can infer a modification date from harvest infos
    change_status, change_payload = await detect_resource_change_from_harvest(data, resource_id)
    if change_status != Change.NO_GUESS:
        return change_status, change_payload

    # if not, let's see if we can infer a modifification date from last-modified headers
    (
        change_status,
        change_payload,
    ) = await detect_resource_change_from_last_modified_header(data)
    if change_status != Change.NO_GUESS:
        return change_status, change_payload

    # if not, let's see if we can infer a modifification date from content-length header
    return await detect_resource_change_from_content_length_header(data)


async def detect_resource_change_from_harvest(
    checks_data: dict, resource_id: str
) -> Tuple[Change, Union[dict, None]]:
    """
    Checks if resource has a harvest.modified_at
    Returns a tuple with a Change status and an optional payload:
    {
        "analysis:last-modified-at": last_modified_date,
        "analysis:last-modified-detection": "harvest-resource-metadata",
    }
    """
    if len(checks_data) == 1:
        return Change.NO_GUESS, None
    last_check = checks_data[1]
    q = """
        SELECT catalog.harvest_modified_at FROM catalog
        WHERE catalog.resource_id = $1
    """
    pool = await context.pool()
    async with pool.acquire() as connection:
        catalog_data = await connection.fetchrow(q, resource_id)
        if catalog_data and catalog_data["harvest_modified_at"]:
            if catalog_data["harvest_modified_at"] == last_check["detected_last_modified_at"]:
                return Change.HAS_NOT_CHANGED, None
            return Change.HAS_CHANGED, {
                "analysis:last-modified-at": catalog_data["harvest_modified_at"].isoformat(),
                "analysis:last-modified-detection": "harvest-resource-metadata",
            }
    return Change.NO_GUESS, None



from datetime import date, datetime
from typing import Union

from dateparser import parse as date_parser
from dateutil.parser import ParserError
from dateutil.parser import parse as dateutil_parser


def to_json(value: str) -> str:
    """Convenience method, should be casted from string directly by postgres"""
    return value


def _parse_dt(value: str) -> Union[datetime, None]:
    """For performance reasons, we try first with dateutil and fallback on dateparser"""
    try:
        return dateutil_parser(value)
    except ParserError:
        return date_parser(value)


def to_date(value: str) -> Union[date, None]:
    parsed = _parse_dt(value)
    return parsed.date() if parsed else None


def to_datetime(value: str) -> Union[datetime, None]:
    return _parse_dt(value)
